from typing import Generic, TypeVar

from rainbowadn.collection.linear import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *

__all__ = ('StackBridge',)

ElementT = TypeVar('ElementT')
Out = TypeVar('Out')


class StackBridge(
    Reducer[SequenceDispatcher[HashPoint[ElementT], Out], Out],
    Generic[Out, ElementT]
):
    def __init__(self, target: HashPoint[Stack[ElementT]]):
        assert isinstance(target, HashPoint)
        self.target = target

    async def reduce(self, reduce: Reduce[SequenceDispatcher[HashPoint[ElementT], Out], Out]) -> Out:
        resolved: Stack[ElementT] = await self.target.resolve()
        out: Out = reduce.initial
        out = await reduce.reduce(out, LastDispatcher(resolved.element))
        deepest: HashPoint[ElementT] = resolved.element
        for element in [x async for x in Stack.iter(resolved.previous)]:
            assert isinstance(element, HashPoint)
            out = await reduce.reduce(out, PairDispatcher(element, deepest))
            deepest = element
        out = await reduce.reduce(out, FirstDispatcher(deepest))
        return out

    def loose(self) -> Reducer[SequenceDispatcher[HashPoint[ElementT], Out], Out]:
        return self
